from tabulate import tabulate
from functools import reduce

from waiter import terminal
from waiter.data_format import display_data
from waiter.format import format_field_name, format_mem_field, format_timestamp_string

from waiter.display import tabulate_token_services
from waiter.querying import get_target_cluster_from_token, print_no_data, query_token
from waiter.util import guard_no_cluster


def tabulate_token(cluster_or_cluster_group_name, token, token_name, services, token_etag):
    """Given a token, returns a string containing tables for the fields"""
    table = [['Owner', token['owner']]]
    if token.get('name'):
        table.append(['Name', token['name']])
    if token.get('cluster'):
        table.append(['Primary Cluster', token['cluster']])
    if token.get('cpus'):
        table.append(['CPUs', token['cpus']])
    if token.get('mem'):
        table.append(['Memory', format_mem_field(token)])
    if token.get('ports'):
        table.append(['Ports requested', token['ports']])
    if token.get('cmd-type'):
        table.append(['Command type', token['cmd-type']])
    if token.get('health-check-url'):
        table.append(['Health check endpoint', token['health-check-url']])
    if token.get('permitted-user'):
        table.append(['Permitted user(s)', token['permitted-user']])

    explicit_keys = ('cluster', 'cmd', 'cmd-type', 'cpus', 'env', 'health-check-url', 'last-update-time',
                     'last-update-user', 'mem', 'name', 'owner', 'permitted-user', 'ports')
    ignored_keys = ('previous', 'root')
    for key, value in token.items():
        if key not in (explicit_keys + ignored_keys):
            table.append([format_field_name(key), value])

    command = token.get('cmd')
    if command:
        token_command = f'Command:\n{command}'
    else:
        token_command = '<No command specified>'

    if token.get('env') and len(token['env']) > 0:
        environment = '\n\nEnvironment:\n%s' % '\n'.join(['%s=%s' % (k, v) for k, v in token['env'].items()])
    else:
        environment = ''

    table_text = tabulate(table, tablefmt='plain')
    last_update_time = format_timestamp_string(token['last-update-time'])
    last_update_user = f' ({token["last-update-user"]})' if 'last-update-user' in token else ''
    column_names = ['Service Id', 'Cluster', 'Run as user', 'Instances', 'CPUs', 'Memory', 'Version', 'Status', 'Last request',
                    'Current?', 'Scaling State']
    service_table, _ = tabulate_token_services(services, token_name, token_etag=token_etag, column_names=column_names)
    return f'\n' \
        f'=== {terminal.bold(cluster_or_cluster_group_name)} / {terminal.bold(token_name)} ===\n' \
        f'\n' \
        f'Last Updated: {last_update_time}{last_update_user}\n' \
        f'\n' \
        f'{table_text}\n' \
        f'\n' \
        f'{token_command}' \
        f'{environment}' \
        f'{service_table}'


def show(clusters, args, _, enforce_cluster):
    """Prints info for the token with the given token name."""
    guard_no_cluster(clusters)
    as_json = args.get('json')
    as_yaml = args.get('yaml')
    token_name = args.get('token')[0]
    include_services = not args.get('no-services')

    query_result = query_token(clusters, token_name, include_services=include_services)
    if as_json or as_yaml:
        display_data(args, query_result)
    elif enforce_cluster:
        for cluster_name, entities in query_result['clusters'].items():
            services = [{'cluster': cluster_name, **service}
                        for service in entities.get('services', [])]
            print(tabulate_token(cluster_name, entities['token'], token_name, services, entities['etag']))
            print()
    else:
        clusters_in_result = [cluster
                              for cluster in clusters
                              if cluster['name'] in query_result['clusters']]
        def add_cluster_to_cluster_group_dict(cluster_groups, cluster):
            cluster_name = cluster['name']
            cluster_group_name = cluster.get('sync-group', cluster_name)
            clusters_in_group = [*cluster_groups.get(cluster_group_name, []), cluster]
            return {**cluster_groups, f'{cluster_group_name}': clusters_in_group}
        cluster_group_name_to_clusters = reduce(add_cluster_to_cluster_group_dict, clusters_in_result, dict())

        for cluster_group_name, cluster_group_clusters in cluster_group_name_to_clusters.items():
            primary_cluster = get_target_cluster_from_token(cluster_group_clusters, token_name, False)
            entities = query_result['clusters'][primary_cluster['name']]
            all_services_in_cluster_group = [{'cluster': cluster['name'], **service}
                                             for cluster in cluster_group_clusters
                                             for service in query_result['clusters'][cluster['name']].get('services', [])]
            print(tabulate_token(cluster_group_name, entities['token'], token_name, all_services_in_cluster_group, entities['etag']))
            print()

    if query_result['count'] > 0:
        return 0
    else:
        if not as_json and not as_yaml:
            print_no_data(clusters)
        return 1


def register(add_parser):
    """Adds this sub-command's parser and returns the action function"""
    show_parser = add_parser('show', help='show token by name')
    show_parser.add_argument('token', nargs=1)
    show_parser.add_argument('--no-services', help="don't show the token's services",
                             dest='no-services', action='store_true')
    format_group = show_parser.add_mutually_exclusive_group()
    format_group.add_argument('--json', help='show the data in JSON format', dest='json', action='store_true')
    format_group.add_argument('--yaml', help='show the data in YAML format', dest='yaml', action='store_true')
    return show



import json
import logging
import requests
from urllib.parse import urljoin

from tabulate import tabulate

from waiter import http_util, terminal
from waiter.format import format_last_request_time
from waiter.format import format_status
from waiter.querying import get_service, get_services_using_token
from waiter.querying import print_no_data, query_service, query_services, query_token
from waiter.util import is_service_current, str2bool, response_message, print_error, wait_until


def ping_on_cluster(cluster, timeout, wait_for_request, token_name, service_exists_fn):
    """Pings using the given token name (or ^SERVICE-ID#) in the given cluster."""
    cluster_name = cluster['name']

    def perform_ping():
        status = None
        try:
            default_queue_timeout_millis = 300000
            timeout_seconds = timeout if wait_for_request else 5
            timeout_millis = timeout_seconds * 1000
            headers = {
                'X-Waiter-Queue-Timeout': str(max(default_queue_timeout_millis, timeout_millis)),
                'X-Waiter-Token': token_name,
                'X-Waiter-Timeout': str(timeout_millis)
            }
            read_timeout = timeout_seconds if wait_for_request else (timeout_seconds + 5)
            resp = http_util.get(cluster, '/waiter-ping', headers=headers, read_timeout=read_timeout)
            logging.debug(f'Response status code: {resp.status_code}')
            resp_json = resp.json()
            if resp.status_code == 200:
                status = resp_json.get('service-state', {}).get('status')
                ping_response = resp_json['ping-response']
                ping_response_result = ping_response['result']
                if ping_response_result in ['descriptor-error', 'received-response']:
                    ping_response_status = ping_response['status']
                    if ping_response_status == 200:
                        print(terminal.success('Ping successful.'))
                        result = True
                    else:
                        print_error(f'Ping responded with non-200 status {ping_response_status}.')
                        try:
                            ping_response_waiter_error = json.loads(ping_response['body'])['waiter-error']['message']
                            print_error(ping_response_waiter_error)
                        except json.JSONDecodeError:
                            logging.debug('Ping response is not in json format, cannot display waiter-error message.')
                        except KeyError:
                            logging.debug('Ping response body does not contain waiter-error message.')
                        result = False
                elif ping_response_result == 'timed-out':
                    if wait_for_request:
                        print_error('Ping request timed out.')
                        result = False
                    else:
                        logging.debug('ignoring ping request timeout due to --no-wait')
                        result = True
                else:
                    print_error(f'Encountered unknown ping result: {ping_response_result}.')
                    result = False
            else:
                print_error(response_message(resp_json))
                result = False
        except Exception:
            result = False
            message = f'Encountered error while pinging in {cluster_name}.'
            logging.exception(message)
            if wait_for_request:
                print_error(message)

        return result, status

    def check_service_status():
        result = wait_until(service_exists_fn, timeout=timeout, interval=5)
        if result:
            print(f'Service is currently {format_status(result["status"])}.')
            return True
        else:
            print_error('Timeout while waiting for service to start.')
            return False

    succeeded, service_status = perform_ping()
    if succeeded:
        if service_status is None or service_status == 'Inactive':
            check_service_status()
        else:
            print(f'Service is currently {format_status(service_status)}.')
    return succeeded


def token_has_current_service(cluster, token_name, current_token_etag):
    """If the given token has a "current" service, returns that service else None"""
    services = get_services_using_token(cluster, token_name)
    if services is not None:
        services = [s for s in services if is_service_current(s, current_token_etag, token_name)]
        return services[0] if len(services) > 0 else False
    else:
        cluster_name = cluster['name']
        print_error(f'Unable to retrieve services using token {token_name} in {cluster_name}.')
        return None


def ping_token_on_cluster(cluster, token_name, timeout, wait_for_request,
                          current_token_etag):
    """Pings the token with the given token name in the given cluster."""
    cluster_name = cluster['name']
    print(f'Pinging token {terminal.bold(token_name)} '
          f'in {terminal.bold(cluster_name)}...')
    return ping_on_cluster(cluster, timeout, wait_for_request,
                           token_name, lambda: token_has_current_service(cluster, token_name, current_token_etag))


def service_is_active(cluster, service_id):
    """If the given service id is active, returns the service, else None"""
    service = get_service(cluster, service_id)
    return service if (service and service.get('status') != 'Inactive') else None


def ping_service_on_cluster(cluster, service_id, timeout, wait_for_request):
    """Pings the service with the given service id in the given cluster."""
    cluster_name = cluster['name']
    print(f'Pinging service {terminal.bold(service_id)} '
          f'in {terminal.bold(cluster_name)}...')
    return ping_on_cluster(cluster, timeout, wait_for_request,
                           f'^SERVICE-ID#{service_id}', lambda: service_is_active(cluster, service_id))


def kill_service_on_cluster(cluster, service_id, timeout_seconds):
    """Kills the service with the given service id in the given cluster."""
    cluster_name = cluster['name']
    http_util.set_retries(0)
    try:
        print(f'Killing service {terminal.bold(service_id)} in {terminal.bold(cluster_name)}...')
        params = {'timeout': timeout_seconds * 1000}
        resp = http_util.delete(cluster, f'/apps/{service_id}', params=params, read_timeout=timeout_seconds)
        logging.debug(f'Response status code: {resp.status_code}')
        if resp.status_code == 200:
            routers_agree = resp.json().get('routers-agree')
            if routers_agree:
                print(f'Successfully killed {service_id} in {cluster_name}.')
                return True
            else:
                print(f'Successfully killed {service_id} in {cluster_name}. '
                      f'Server-side timeout waiting for routers to update.')
                return False
        else:
            print_error(response_message(resp.json()))
            return False
    except requests.exceptions.ReadTimeout:
        message = f'Request timed out while killing {service_id} in {cluster_name}.'
        logging.exception(message)
        print_error(message)
    except Exception:
        message = f'Encountered error while killing {service_id} in {cluster_name}.'
        logging.exception(message)
        print_error(message)


def process_kill_request(clusters, token_name_or_service_id, is_service_id, force_flag, timeout_secs,
                         no_service_result=False):
    """Kills the service(s) using the given token name or service-id.
    Returns False if no services can be found or if there was a failure in deleting any service.
    Returns True if all services using the token were deleted successfully."""
    if is_service_id:
        query_result = query_service(clusters, token_name_or_service_id)
        num_services = query_result['count']
        if num_services == 0:
            print_no_data(clusters)
            return no_service_result
    else:
        query_result = query_services(clusters, token_name_or_service_id)
        num_services = query_result['count']
        if num_services == 0:
            clusters_text = ' / '.join([terminal.bold(c['name']) for c in clusters])
            print(f'There are no services using token {terminal.bold(token_name_or_service_id)} in {clusters_text}.')
            return no_service_result
        elif num_services > 1:
            print(f'There are {num_services} services using token {terminal.bold(token_name_or_service_id)}:')

    cluster_data_pairs = sorted(query_result['clusters'].items())
    clusters_by_name = {c['name']: c for c in clusters}
    overall_success = True
    for cluster_name, data in cluster_data_pairs:
        if is_service_id:
            service = data['service']
            service['service-id'] = token_name_or_service_id
            services = [service]
        else:
            services = sorted(data['services'], key=lambda s: s.get('last-request-time', None) or '', reverse=True)

        for service in services:
            service_id = service['service-id']
            cluster = clusters_by_name[cluster_name]
            status_string = service['status']
            status = format_status(status_string)
            inactive = status_string == 'Inactive'
            should_kill = False
            if force_flag or num_services == 1:
                should_kill = True
            else:
                url = urljoin(cluster['url'], f'apps/{service_id}')
                if is_service_id:
                    active_instances = service['instances']['active-instances']
                    healthy_count = len([i for i in active_instances if i['healthy?']])
                    unhealthy_count = len([i for i in active_instances if not i['healthy?']])
                else:
                    healthy_count = service['instance-counts']['healthy-instances']
                    unhealthy_count = service['instance-counts']['unhealthy-instances']

                last_request_time = format_last_request_time(service)
                run_as_user = service['service-description']['run-as-user']
                table = [['Status', status],
                         ['Healthy', healthy_count],
                         ['Unhealthy', unhealthy_count],
                         ['URL', url],
                         ['Run as user', run_as_user]]
                table_text = tabulate(table, tablefmt='plain')
                print(f'\n'
                      f'=== {terminal.bold(cluster_name)} / {terminal.bold(service_id)} ===\n'
                      f'\n'
                      f'Last Request: {last_request_time}\n'
                      f'\n'
                      f'{table_text}\n')
                if not inactive:
                    answer = input(f'Kill service in {terminal.bold(cluster_name)}? ')
                    should_kill = str2bool(answer)

            if inactive:
                print(f'Service {terminal.bold(service_id)} on {terminal.bold(cluster_name)} cannot be killed because '
                      f'it is already {status}.')
                should_kill = False

            if should_kill:
                success = kill_service_on_cluster(cluster, service_id, timeout_secs)
                overall_success = overall_success and success
    return overall_success


def token_explicitly_created_on_cluster(cluster, token_cluster_name):
    """Returns true if the given token cluster matches the configured cluster name of the given cluster
    or if the token cluster matches the configured cluster name in the .waiter.json file"""
    cluster_settings, _ = http_util.make_data_request(cluster, lambda: http_util.get(cluster, '/settings'))
    cluster_config_name = cluster_settings['cluster-config']['name'].upper()
    cluster_local_config_name = cluster['name'].upper()

    # we consider the token having the same cluster value as the configured cluster in .waiter.json as an alias
    # which makes it easier to change a Waiter cluster name without making breaking changes to the CLI client.
    created_on_this_cluster = token_cluster_name in [cluster_config_name, cluster_local_config_name]
    return created_on_this_cluster


def process_ping_request(clusters, token_name_or_service_id, is_service_id, timeout_secs, wait_for_request):
    """Pings the service using the given token name or service-id.
    Returns False if the ping request fails or times out.
    Returns True if the ping request completes successfully."""
    if is_service_id:
        query_result = query_service(clusters, token_name_or_service_id)
    else:
        query_result = query_token(clusters, token_name_or_service_id)

    if query_result['count'] == 0:
        print_no_data(clusters)
        return False

    http_util.set_retries(0)
    cluster_data_pairs = sorted(query_result['clusters'].items())
    clusters_by_name = {c['name']: c for c in clusters}
    overall_success = True
    for cluster_name, data in cluster_data_pairs:
        cluster = clusters_by_name[cluster_name]
        if is_service_id:
            success = ping_service_on_cluster(cluster, token_name_or_service_id, timeout_secs, wait_for_request)
        else:
            token_data = data['token']
            token_etag = data['etag']
            token_cluster_name = token_data['cluster'].upper()
            if len(clusters) == 1 or token_explicitly_created_on_cluster(cluster, token_cluster_name):
                success = ping_token_on_cluster(cluster, token_name_or_service_id, timeout_secs,
                                                wait_for_request, token_etag)
            else:
                print(f'Not pinging token {terminal.bold(token_name_or_service_id)} '
                      f'in {terminal.bold(cluster_name)} '
                      f'because it was created in {terminal.bold(token_cluster_name)}.')
                success = True
        overall_success = overall_success and success
    return overall_success



import argparse
import logging
import os
from enum import Enum

from waiter import plugins, terminal
from waiter.display import get_user_selection, tabulate_service_instances, tabulate_token_services
from waiter.querying import get_service_id_from_instance_id, get_target_cluster_from_token, print_no_data, \
    print_no_services, query_service, query_token, get_services_on_cluster, print_no_instances
from waiter.util import guard_no_cluster, print_info

BASH_PATH = '/bin/bash'


class Destination(Enum):
    TOKEN = 'token'
    SERVICE_ID = 'service_id'
    INSTANCE_ID = 'instance_id'


def map_instances_with_status(instances, status):
    return [{'_status': status, **inst} for inst in instances]


def get_instances_from_service_id(clusters, service_id, include_active_instances, include_failed_instances,
                                  include_killed_instances):
    query_result = query_service(clusters, service_id)
    num_services = query_result['count']
    if num_services == 0:
        return False
    services = map(lambda service_data: service_data['service'], query_result['clusters'].values())
    instances = []
    for service in services:
        if include_active_instances:
            instances += map_instances_with_status(service['instances']['active-instances'], 'active')
        if include_failed_instances:
            instances += map_instances_with_status(service['instances']['failed-instances'], 'failed')
        if include_killed_instances:
            instances += map_instances_with_status(service['instances']['killed-instances'], 'killed')
    return instances


def kubectl_exec_to_instance(kubectl_cmd, namespace, pod_name, container_name, log_directory,
                             command_to_run=None, api_server=None, context=None):
    if context is not None:
        args = ['--context', context]
    elif api_server is not None:
        args = ['--server', api_server]
    args = [*args,
            '--namespace', namespace,
            'exec',
            '-it', pod_name,
            '-c', container_name,
            '--',
            '/bin/bash', '-c', f"cd {log_directory}; {' '.join(command_to_run) or 'exec /bin/bash'}"]
    os.execlp(kubectl_cmd, 'kubectl', *args)


def ssh_instance(instance, container_name, command_to_run=None):
    print_info(f'Attempting to ssh into instance {terminal.bold(instance["id"])}...')
    log_directory = instance['log-directory']
    k8s_pod_name = instance.get('k8s/pod-name', False)
    if k8s_pod_name:
        context = instance.get('k8s/context')
        k8s_api_server = instance['k8s/api-server-url']
        kubectl_cmd = os.getenv('WAITER_KUBECTL', plugins.get_fn('get-kubectl-cmd', lambda: 'kubectl')())
        k8s_namespace = instance['k8s/namespace']
        print_info(f'Executing ssh to k8s pod {terminal.bold(k8s_pod_name)}')
        logging.debug(f'Executing ssh to k8s pod {terminal.bold(k8s_pod_name)} '
                      f'using namespace={k8s_namespace} api_server={k8s_api_server}')
        kubectl_exec_to_instance(kubectl_cmd, k8s_namespace, k8s_pod_name, container_name,
                                 log_directory, command_to_run, api_server=k8s_api_server, context=context)
    else:
        hostname = instance['host']
        command_to_run = command_to_run or [BASH_PATH]
        ssh_cmd = os.getenv('WAITER_SSH', 'ssh')
        args = [ssh_cmd, '-t', hostname, 'cd', log_directory, ';'] + command_to_run
        print_info(f'Executing ssh to {terminal.bold(hostname)}')
        os.execlp(ssh_cmd, *args)


def ssh_instance_id(clusters, instance_id, command, container_name):
    service_id = get_service_id_from_instance_id(instance_id)
    instances = get_instances_from_service_id(clusters, service_id, True, True, True)
    if instances is False:
        print_no_data(clusters)
        return 1
    found_instance = next((instance
                           for instance in instances
                           if instance['id'] == instance_id),
                          False)
    if not found_instance:
        print_no_data(clusters)
        return 1
    return ssh_instance(found_instance, container_name, command)


def ssh_service_id(clusters, service_id, command, container_name, skip_prompts, include_active_instances,
                   include_failed_instances, include_killed_instances):
    instances = get_instances_from_service_id(clusters, service_id, include_active_instances, include_failed_instances,
                                              include_killed_instances)
    if instances is False:
        print_no_data(clusters)
        return 1
    if len(instances) == 0:
        print_no_instances(service_id)
        return 1
    if skip_prompts:
        selected_instance = instances[0]
    else:
        column_names = ['Instance Id', 'Host', 'Status']
        tabular_output = tabulate_service_instances(instances, show_index=True, column_names=column_names)
        selected_instance = get_user_selection(instances, tabular_output)
    return ssh_instance(selected_instance, container_name, command)


def ssh_token(clusters, enforce_cluster, token, command, container_name, skip_prompts, include_active_instances,
              include_failed_instances, include_killed_instances):
    if skip_prompts:
        cluster = get_target_cluster_from_token(clusters, token, enforce_cluster)
        query_result = get_services_on_cluster(cluster, token)
        services = [s
                    for s in query_result.get('services', [])
                    if s['instance-counts']['healthy-instances'] + s['instance-counts']['unhealthy-instances'] > 0]
        if len(services) == 0:
            print_no_services(clusters, token)
            return 1
        max_last_request = max(s.get('last-request-time', '') for s in services)
        selected_service_id = next(s['service-id'] for s in services if s['last-request-time'] == max_last_request)
    else:
        query_result = query_token(clusters, token, include_services=True)
        if query_result['count'] == 0:
            print_no_data(clusters)
            return 1
        clusters_by_name = {c['name']: c for c in clusters}
        cluster_data = query_result['clusters']
        services = [{'cluster': cluster, 'etag': data['etag'], **service}
                    for cluster, data in cluster_data.items()
                    for service in data['services']]
        if len(services) == 0:
            print_no_services(clusters, token)
            return 1
        column_names = ['Service Id', 'Cluster', 'Instances', 'In-flight req.', 'Status', 'Last request', 'Current?']
        tabular_output, sorted_services = tabulate_token_services(services, token, show_index=True, summary_table=False,
                                                                  column_names=column_names)
        selected_service = get_user_selection(sorted_services, tabular_output)
        selected_service_id = selected_service['service-id']
        clusters = [clusters_by_name[selected_service['cluster']]]
    return ssh_service_id(clusters, selected_service_id, command, container_name, skip_prompts,
                          include_active_instances, include_failed_instances, include_killed_instances)


def ssh(clusters, args, _, enforce_cluster):
    guard_no_cluster(clusters)
    token_or_service_id_or_instance_id = args.pop('token-or-service-id-or-instance-id')
    command = args.pop('command')
    ssh_destination = args.pop('ssh_destination')
    include_active_instances = args.pop('include_active_instances')
    include_failed_instances = args.pop('include_failed_instances')
    include_killed_instances = args.pop('include_killed_instances')
    container_name = args.pop('container_name', 'waiter-app')
    skip_prompts = args.pop('quick')
    if ssh_destination == Destination.TOKEN:
        return ssh_token(clusters, enforce_cluster, token_or_service_id_or_instance_id, command, container_name,
                         skip_prompts, include_active_instances, include_failed_instances, include_killed_instances)
    elif ssh_destination == Destination.SERVICE_ID:
        return ssh_service_id(clusters, token_or_service_id_or_instance_id, command, container_name, skip_prompts,
                              include_active_instances, include_failed_instances, include_killed_instances)
    elif ssh_destination == Destination.INSTANCE_ID:
        return ssh_instance_id(clusters, token_or_service_id_or_instance_id, command, container_name)


def register(add_parser):
    """Adds this sub-command's parser and returns the action function"""
    parser = add_parser('ssh',
                        help='ssh to a Waiter instance',
                        description='ssh to an instance given the token, service id, or instance id. Working directory '
                                    'will be the log directory.')
    parser.add_argument('token-or-service-id-or-instance-id')
    parser.add_argument('--container-name', '-c',
                        help='specify the container name you want to ssh into. Defaults to "waiter-app". Has no '
                                'effect if instance is not k8s pod.')
    id_group = parser.add_mutually_exclusive_group(required=False)
    id_group.add_argument('--token', '-t', dest='ssh_destination', action='store_const', const=Destination.TOKEN,
                          default=Destination.TOKEN, help='Default; ssh with token')
    id_group.add_argument('--service-id', '-s', dest='ssh_destination', action='store_const',
                          const=Destination.SERVICE_ID, help='ssh using a service id')
    id_group.add_argument('--instance-id', '-i', dest='ssh_destination', action='store_const',
                          const=Destination.INSTANCE_ID, help='ssh directly to instance id')
    parser.add_argument('--quick', '-q', dest='quick', action='store_true',
                        help='skips services prompt by selecting the service with latest request, and instances prompt '
                             'by selecting a random one.')
    active_group = parser.add_mutually_exclusive_group(required=False)
    failed_group = parser.add_mutually_exclusive_group(required=False)
    killed_group = parser.add_mutually_exclusive_group(required=False)
    active_group.add_argument('--active', '-a', dest='include_active_instances', action='store_true', default=True,
                              help='included by default; includes active instances when prompting')
    failed_group.add_argument('--failed', '-f', dest='include_failed_instances', action='store_true',
                              help='includes failed instances when prompting')
    killed_group.add_argument('--killed', '-k', dest='include_killed_instances', action='store_true',
                              help='includes killed instances when prompting')
    active_group.add_argument('--no-active', dest='include_active_instances', action='store_false',
                              help="don't show active instances in prompt")
    failed_group.add_argument('--no-failed', dest='include_failed_instances', action='store_false',
                              help="don't show failed instances in prompt")
    killed_group.add_argument('--no-killed', dest='include_killed_instances', action='store_false',
                              help="don't show killed instances in prompt")
    parser.add_argument('command', nargs=argparse.REMAINDER, help='command to be run on instance')
    return ssh



import importlib
import json
import logging
import uuid
from urllib.parse import urljoin

import requests

import waiter
from waiter.util import print_error


session = None
timeouts = None
adapter_factory = None


def set_retries(retries):
    """Sets the number of retries to use"""
    adapter = adapter_factory(max_retries=retries)
    session.mount('http://', adapter)
    session.mount('https://', adapter)


def configure(config, plugins):
    """Configures HTTP timeouts and retries to be used"""
    global session
    global timeouts
    global adapter_factory
    adapter_factory = plugins.get('http-adapter-factory', requests.adapters.HTTPAdapter)
    session_factory = plugins.get('http-session-factory', requests.Session)
    logging.getLogger('urllib3').setLevel(logging.DEBUG) # logging.disable in cli.py may override
    http_config = config.get('http')
    connect_timeout = http_config.get('connect-timeout')
    read_timeout = http_config.get('read-timeout')
    timeouts = (connect_timeout, read_timeout)
    logging.debug('using http timeouts: %s', timeouts)
    retries = http_config.get('retries')
    session = session_factory()
    set_retries(retries)
    session.headers['User-Agent'] = f"waiter/{waiter.version.VERSION} ({session.headers['User-Agent']})"
    auth_config = http_config.get('auth', None)
    if auth_config:
        auth_type = auth_config.get('type')
        if auth_type == 'basic':
            basic_auth_config = auth_config.get('basic')
            user = basic_auth_config.get('user')
            session.auth = (user, basic_auth_config.get('pass'))
            logging.debug(f'using http basic auth with user {user}')
        else:
            raise Exception(f'Encountered unsupported authentication type "{auth_type}".')


def __post(url, json_body, params=None, **kwargs):
    """Sends a POST with the json payload to the given url"""
    logging.debug(f'POST {url} with body {json_body} and headers {kwargs.get("headers", {})}')
    return session.post(url, json=json_body, timeout=timeouts, params=params, **kwargs)


def __get(url, params=None, read_timeout=None, **kwargs):
    """Sends a GET with params to the given url"""
    logging.debug(f'GET {url} with params {params} and headers {kwargs.get("headers", {})}')
    get_timeouts = timeouts
    if read_timeout is not None:
        get_timeouts = (timeouts[0], read_timeout)
    return session.get(url, params=params, timeout=get_timeouts, **kwargs)


def __delete(url, params=None, headers=None, read_timeout=None):
    """Sends a DELETE with params to the given url"""
    logging.debug(f'DELETE {url} with params {params} and headers {headers}')
    delete_timeouts = timeouts
    if read_timeout is not None:
        delete_timeouts = (timeouts[0], read_timeout)
    return session.delete(url, params=params, timeout=delete_timeouts, headers=headers)


def __make_url(cluster, endpoint):
    """Given a cluster and an endpoint, returns the corresponding full URL"""
    return urljoin(cluster['url'], endpoint)


def default_http_headers():
    """Returns the default HTTP headers, including a random CID in x-cid"""
    return {
        'Accept': 'application/json',
        'x-cid': f'waiter-{waiter.version.VERSION}-{uuid.uuid4()}'
    }


def post(cluster, endpoint, json_body, params=None, headers=None):
    """POSTs data to cluster at /endpoint"""
    if headers is None:
        headers = {}
    url = __make_url(cluster, endpoint)
    default_headers = default_http_headers()
    resp = __post(url, json_body, params=params, headers={**default_headers, **headers})
    resp.headers.pop('Set-Cookie', None)
    logging.info(f'POST response: {resp.text} (headers: {resp.headers})')
    return resp


def get(cluster, endpoint, params=None, headers=None, read_timeout=None):
    """GETs data corresponding to the given params from cluster at /endpoint"""
    if headers is None:
        headers = {}
    url = __make_url(cluster, endpoint)
    default_headers = default_http_headers()
    resp = __get(url, params, headers={**default_headers, **headers}, read_timeout=read_timeout)
    resp.headers.pop('Set-Cookie', None)
    logging.info(f'GET response: {resp.text} (headers: {resp.headers})')
    return resp


def delete(cluster, endpoint, params=None, headers=None, read_timeout=None):
    """DELETEs data corresponding to the given params on cluster at /endpoint"""
    if headers is None:
        headers = {}
    url = __make_url(cluster, endpoint)
    default_headers = default_http_headers()
    resp = __delete(url, params, headers={**default_headers, **headers}, read_timeout=read_timeout)
    logging.info(f'DELETE response: {resp.text}')
    return resp


def make_data_request(cluster, make_request_fn):
    """
    Makes a request (using make_request_fn), parsing the
    assumed-to-be-JSON response and handling common errors
    """
    try:
        resp = make_request_fn()
        if resp.status_code == 200:
            return resp.json(), resp.headers
        elif resp.status_code == 401:
            print_error(f'Authentication failed on {cluster["name"]} ({cluster["url"]}).')
            return [], {}
        elif resp.status_code == 500:
            print_error(f'Encountered server error while querying {cluster["name"]}.')
            # fall through to logging call below

        logging.warning(f'Unexpected response code {resp.status_code} for data request. Response body: {resp.text}')
    except requests.exceptions.ConnectionError as ce:
        logging.exception(ce)
        print_error(f'Encountered connection error with {cluster["name"]} ({cluster["url"]}).')
    except requests.exceptions.ReadTimeout as rt:
        logging.exception(rt)
        print_error(f'Encountered read timeout with {cluster["name"]} ({cluster["url"]}).')
    except IOError as ioe:
        logging.exception(ioe)
    except json.decoder.JSONDecodeError as jde:
        logging.exception(jde)
    return None, {}



import argparse
import logging
from urllib.parse import urlparse

from waiter import configuration, http_util, metrics, version
from waiter.subcommands import create, delete, init, kill, maintenance, ping, show, ssh, tokens, update
import waiter.plugins as waiter_plugins

parser = argparse.ArgumentParser(description='waiter is the Waiter CLI')
parser.add_argument('--cluster', '-c', help='the name of the Waiter cluster to use')
parser.add_argument('--url', '-u', help='the url of the Waiter cluster to use')
parser.add_argument('--config', '-C', help='the configuration file to use')
parser.add_argument('--verbose', '-v', help='be more verbose/talkative (useful for debugging)',
                    dest='verbose', action='store_true')
parser.add_argument('--version', help='output version information and exit',
                    version=f'%(prog)s version {version.VERSION}', action='version')

subparsers = parser.add_subparsers(dest='action')

actions = {
    'create': {
        'run-function': create.register(subparsers.add_parser),
        'implicit-args-function': create.add_implicit_arguments
    },
    'delete': {
        'run-function': delete.register(subparsers.add_parser)
    },
    'init': {
        'run-function': init.register(subparsers.add_parser),
        'implicit-args-function': init.add_implicit_arguments
    },
    'kill': {
        'run-function': kill.register(subparsers.add_parser)
    },
    'maintenance': {
        'run-function': maintenance.register(subparsers.add_parser)
    },
    'ping': {
        'run-function': ping.register(subparsers.add_parser)
    },
    'show': {
        'run-function': show.register(subparsers.add_parser)
    },
    'ssh': {
        'run-function': ssh.register(subparsers.add_parser)
    },
    'start': {
        'run-function': maintenance.register_stop('start', subparsers.add_parser)
    },
    'stop': {
        'run-function': maintenance.register_start('stop', subparsers.add_parser)
    },
    'tokens': {
        'run-function': tokens.register(subparsers.add_parser)
    },
    'update': {
        'run-function': update.register(subparsers.add_parser),
        'implicit-args-function': update.add_implicit_arguments
    }
}


def load_target_clusters(config_map, url=None, cluster=None):
    """Given the config and (optional) url and cluster flags, returns the list of clusters to target"""
    if cluster and url:
        raise Exception('You cannot specify both a cluster name and a cluster url at the same time')

    clusters = None
    config_clusters = config_map.get('clusters')
    if url:
        if urlparse(url).scheme == '':
            url = 'http://%s' % url
        clusters = [{'name': url, 'url': url}]
    elif config_clusters:
        if cluster:
            clusters = [c for c in config_clusters if c.get('name').lower() == cluster.lower()]
            if len(clusters) == 0 and len(config_clusters) > 0:
                config_cluster_names = ', '.join([c.get('name') for c in config_clusters])
                raise Exception(f'You specified cluster {cluster}, which was not present in your config.' +
                                f' You have the following clusters configured: {config_cluster_names}.')
        else:
            clusters = [c for c in config_clusters if 'disabled' not in c or not c['disabled']]

    return clusters


def run(args, plugins):
    """
    Main entrypoint to the Waiter CLI. Loads configuration files,
    processes global command line arguments, and calls other command line 
    sub-commands (actions) if necessary.
    """
    args, unknown_args = parser.parse_known_args(args)
    verbose = args.verbose
    if verbose:
        log_format = '%(asctime)s [%(levelname)s] [%(name)s] %(message)s'
        logging.getLogger('').handlers = []
        logging.basicConfig(format=log_format, level=logging.DEBUG)
    else:
        logging.disable(logging.FATAL)

    if args.action:
        add_implicit_arguments = actions[args.action].get('implicit-args-function', None)
        if add_implicit_arguments:
            add_implicit_arguments(unknown_args)

    args = parser.parse_args()
    args = vars(args)
    logging.debug('args: %s', args)
    args.pop('verbose')

    action = args.pop('action')
    config_path = args.pop('config')
    cluster = args.pop('cluster')
    url = args.pop('url')

    logging.debug('plugins: %s', plugins)
    waiter_plugins.configure(plugins)

    if action is None:
        parser.print_help()
    else:
        config_map = configuration.load_config_with_defaults(config_path)
        try:
            metrics.initialize(config_map)
            metrics.inc(f'command.{action}.runs')
            clusters = load_target_clusters(config_map, url, cluster)
            enforce_cluster = (url or cluster) and True
            http_util.configure(config_map, plugins)
            args = {k: v for k, v in args.items() if v is not None}
            result = actions[action]['run-function'](clusters, args, config_path, enforce_cluster)
            logging.debug(f'result: {result}')
            if result == 0:
                metrics.inc(f'command.{action}.result.success')
            else:
                metrics.inc(f'command.{action}.result.failure')
            return result
        finally:
            metrics.close()

    return None




from waiter.action import process_ping_request
from waiter.util import check_positive, guard_no_cluster


def ping(clusters, args, _, __):
    """Pings the token with the given token name."""
    guard_no_cluster(clusters)
    token_name_or_service_id = args.get('token-or-service-id')
    is_service_id = args.get('is-service-id', False)
    timeout_secs = args.get('timeout', None)
    wait_for_request = args.get('wait', True)
    ping_success = process_ping_request(clusters, token_name_or_service_id, is_service_id,
                                        timeout_secs, wait_for_request)
    return 0 if ping_success else 1


def register(add_parser):
    """Adds this sub-command's parser and returns the action function"""
    default_timeout = 300
    parser = add_parser('ping', help='ping token by name')
    parser.add_argument('token-or-service-id')
    parser.add_argument('--timeout', '-t', help=f'read timeout (in seconds) for ping request (default is '
                                                f'{default_timeout} seconds)',
                        type=check_positive, default=default_timeout)
    parser.add_argument('--service-id', '-s', help='ping by service id instead of token',
                        dest='is-service-id', action='store_true')
    parser.add_argument('--no-wait', '-n', help='do not wait for ping request to return',
                        dest='wait', action='store_false')
    return ping



__plugins = {}


def configure(plugins):
    """Configures global plugins to the plugins map"""
    global __plugins
    __plugins = plugins


def get_fn(plugin_name, default_fn):
    """Returns the plugin function corresponding to the given plugin name if found, otherwise, default_fn"""
    return __plugins.get(plugin_name, default_fn)



import os
import sys

import textwrap


MOVE_UP = '\033[F'


class Color:
    PURPLE = '\033[95m'
    CYAN = '\033[96m'
    DARKCYAN = '\033[36m'
    BLUE = '\033[94m'
    GREEN = '\033[92m'
    YELLOW = '\033[93m'
    RED = '\033[91m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    END = '\033[0m'


def failed(s):
    return colorize(s, Color.BOLD + Color.RED)


def success(s):
    return colorize(s, Color.GREEN)


def running(s):
    return colorize(s, Color.CYAN)


def inactive(s):
    return colorize(s, Color.YELLOW)


def starting(s):
    return colorize(s, Color.BLUE)


def reason(s):
    return colorize(s, Color.RED)


def bold(s):
    return colorize(s, Color.BOLD)


wrap = textwrap.wrap


def colorize(s, color):
    """Formats the given string with the given color"""
    return f'{color}{s}{Color.END}' if tty() else s


def tty():
    """Returns true if running in a real terminal (as opposed to being piped or redirected). If WAITER_TTY is set then
    return True else False"""
    return os.getenv("WAITER_FORCE_CLI_TTY", False) or sys.stdout.isatty()



import collections

from tabulate import tabulate

from waiter import terminal
from waiter.format import format_last_request_time, format_mem_field, format_memory_amount, format_status
from waiter.util import is_service_current, print_error


def retrieve_num_instances(service):
    """Returns the total number of instances."""
    instance_counts = service["instance-counts"]
    return instance_counts["healthy-instances"] + instance_counts["unhealthy-instances"]


def format_using_current_token(service, token_etag, token_name):
    """Formats the "Current?" column for the given service"""
    is_current = is_service_current(service, token_etag, token_name)
    if is_current:
        return terminal.success('Current')
    else:
        return 'Not Current'


def format_instance_status(instance):
    """Formats the "Healthy?" column for the given instance"""
    if instance['healthy?']:
        return terminal.success('Healthy')
    else:
        if instance['_status'] == 'failed':
            status = 'Failed'
        elif instance['_status'] == 'killed':
            status = 'Killed'
        else:
            status = 'Unhealthy'
        return terminal.failed(status)


def tabulate_token_services(services, token_name, token_etag=None, show_index=False, summary_table=True,
                            column_names=[]):
    """
    :param services: list of services to be displayed as rows
    :param token_name: the token that the services belong to
    :param token_etag: token_etag determines if a the service is current, will default to service['etag'] or None
    :param show_index: shows index column (usually for future choice prompt)
    :param summary_table: provides summary table of services (e.g. Total Memory, CPUS)
    :param column_names: column fields to be included in table
    :return: tabular output string and sorted services list in descending order by last request time
    """
    num_services = len(services)
    if num_services > 0:
        services = sorted(services, key=lambda s: s.get('last-request-time', None) or '', reverse=True)
        rows = [collections.OrderedDict([(key, data)
                                         for key, data in
                                         [('Index', f'[{index + 1}]'),
                                          ('Service Id', s['service-id']),
                                          ('Cluster', s.get('cluster', None)),
                                          ('Run as user', s['effective-parameters']['run-as-user']),
                                          ('Instances', retrieve_num_instances(s)),
                                          ('CPUs', s['effective-parameters']['cpus']),
                                          ('Memory', format_mem_field(s['effective-parameters'])),
                                          ('Version', s['effective-parameters']['version']),
                                          ('In-flight req.', s['request-metrics']['outstanding']),
                                          ('Status', format_status(s['status'])),
                                          ('Last request', format_last_request_time(s)),
                                          ('Current?',
                                           format_using_current_token(s, token_etag or s.get('etag', None),
                                                                      token_name)),
                                          ('Scaling State', s.get('scaling-state', 'Unknown'))]
                                         if key in column_names or show_index and key == 'Index'])
                for index, s in enumerate(services)]
        service_table = tabulate(rows, headers='keys', tablefmt='plain')
        if summary_table:
            num_failing_services = len([s for s in services if s['status'] == 'Failing'])
            num_instances = sum(retrieve_num_instances(s) for s in services)
            total_mem_usage = format_memory_amount(sum(s['resource-usage']['mem'] for s in services))
            total_cpu_usage = round(sum(s['resource-usage']['cpus'] for s in services), 2)
            table = [['# Services', num_services],
                     ['# Failing', num_failing_services],
                     ['# Instances', num_instances],
                     ['Total Memory', total_mem_usage],
                     ['Total CPUs', total_cpu_usage]]
            summary_table = tabulate(table, tablefmt='plain')
            return f'\n\n{summary_table}\n\n{service_table}', services
        else:
            return service_table, services
    else:
        return '', services


def tabulate_service_instances(instances, show_index=False, column_names=[]):
    """
    :param instances: list of instances to be displayed
    :param show_index: shows index column (usually for future choice prompt)
    :param column_names: column names to be displayed in table
    :return: tabular output string
    """
    if len(instances) > 0:
        rows = [collections.OrderedDict([(key, data)
                                         for key, data in
                                         [('Index', f'[{index + 1}]'),
                                          ('Instance Id', inst['id']),
                                          ('Host', inst['host']),
                                          ('Status', format_instance_status(inst))]
                                         if key in column_names or show_index and key == 'Index'])
                for index, inst in enumerate(instances)]
        return tabulate(rows, headers='keys', tablefmt='plain')
    else:
        return ''


def get_user_selection(items, tabular_str, short_circuit_choice=True):
    """
    :param items: list of possible choices
    :param tabular_str: table output to provide user with options
    :param short_circuit_choice: When True and only one item in items, return that item as the selection without user
    prompt.
    :exception Raises exception when user input is invalid
    :return selected item (an element from the items list)
    """
    if short_circuit_choice and len(items) == 1:
        return items[0]
    print(tabular_str)
    answer = input(f'Enter the Index of your choice: ')
    print()
    try:
        index = int(answer) - 1
        if index < 0 or index >= len(items):
            raise Exception('Input is out of range!')
        return items[index]
    except ValueError as error:
        print_error('Input received did not match any of the choices!')
        raise error



import logging

from waiter import http_util, terminal
from waiter.querying import get_services_using_token, get_target_cluster_from_token, print_no_data, query_token
from waiter.util import get_in, guard_no_cluster, response_message, print_error, str2bool


def delete_token_on_cluster(cluster, token_name, token_etag):
    """Deletes the token with the given token name in the given cluster."""
    cluster_name = cluster['name']
    try:
        # Retrieve all services that are using the token
        services = get_services_using_token(cluster, token_name)
        if services is not None:
            num_services = len(services)
            if num_services > 0:
                if num_services == 1:
                    print_error(f'There is one service using token {token_name} in {cluster_name}:\n')
                else:
                    print_error(f'There are {num_services} services using token {token_name} in {cluster_name}:\n')
                for service in services:
                    print_error(f'  {service["url"]}')
                if num_services == 1:
                    print_error('\nPlease kill this service before deleting the token it relies on.')
                else:
                    print_error('\nPlease kill these services before deleting the token they rely on.')
                return False
        else:
            print_error(f'Unable to retrieve services using token {token_name} in {cluster_name}.')
            return False

        # Delete the token
        print(f'Deleting token {terminal.bold(token_name)} in {terminal.bold(cluster_name)}...')
        headers = {
            'If-Match': token_etag,
            'X-Waiter-Token': token_name
        }
        resp = http_util.delete(cluster, '/token', headers=headers)
        logging.debug(f'Response status code: {resp.status_code}')
        if resp.status_code == 200:
            print(f'Successfully deleted {token_name} in {cluster_name}.')
            return True
        else:
            print_error(response_message(resp.json()))
            return False
    except Exception as e:
        message = f'Encountered error while deleting token {token_name} in {cluster_name}.'
        logging.error(e, message)
        print_error(message)


def delete(clusters, args, _, enforce_cluster):
    """Deletes the token with the given token name."""
    guard_no_cluster(clusters)
    force_delete = args.get('force', False)
    token_name = args.get('token')[0]
    query_result = query_token(clusters, token_name)
    if query_result['count'] == 0:
        print_no_data(clusters)
        return 1

    cluster_data_pairs = sorted(query_result['clusters'].items())
    num_clusters = len(cluster_data_pairs)
    clusters_by_name = {c['name']: c for c in clusters}
    sync_opt_out = any(get_in(token_config, ['token', 'metadata', 'waiter-token-sync-opt-out']) == 'true'
                       for _, token_config in cluster_data_pairs)

    if force_delete or sync_opt_out:
        cluster_data_pairs = sorted(query_result['clusters'].items())
        clusters_by_name = {c['name']: c for c in clusters}
        num_clusters = len(cluster_data_pairs)
        cluster_names_found = [p[0] for p in cluster_data_pairs]

        print(f'Token {terminal.bold(token_name)} exists in {num_clusters} cluster(s): {", ".join(cluster_names_found)}.')
        overall_success = True
        for cluster_name, data in cluster_data_pairs:
            if force_delete or (num_clusters == 1):
                should_delete = True
            else:
                should_delete = str2bool(input(f'Delete token in {terminal.bold(cluster_name)}? '))
            if should_delete:
                cluster = clusters_by_name[cluster_name]
                success = delete_token_on_cluster(cluster, token_name, data['etag'])
                overall_success = overall_success and success
        return 0 if overall_success else 1
    else:
        cluster = get_target_cluster_from_token(clusters, token_name, enforce_cluster)
        if cluster is None:
            print_no_data(clusters)
            return 1
        else:
            token_config = query_result['clusters'][cluster['name']]
            token_etag = token_config['etag']
            success = delete_token_on_cluster(cluster, token_name, token_etag)
            return 0 if success else 1


def register(add_parser):
    """Adds this sub-command's parser and returns the action function"""
    parser = add_parser('delete', help='delete token by name')
    parser.add_argument('token', nargs=1)
    parser.add_argument('--force', '-f', help='delete on all clusters where present, never prompt',
                        dest='force', action='store_true')
    return delete



import argparse
import logging
from collections import defaultdict
from enum import Enum

import requests

from waiter import terminal, http_util
from waiter.data_format import determine_format, display_data, load_data
from waiter.querying import get_token, query_token, get_target_cluster_from_token
from waiter.util import deep_merge, FALSE_STRINGS, is_admin_enabled, print_info, response_message, TRUE_STRINGS, \
    guard_no_cluster, str2bool, update_in

BOOL_STRINGS = TRUE_STRINGS + FALSE_STRINGS
INT_PARAM_SUFFIXES = ['-failures', '-index', '-instances', '-length', '-level', '-mins', '-ms', '-secs']
FLOAT_PARAM_SUFFIXES = ['-factor', '-rate', '-threshold']
STRING_PARAM_PREFIXES = ['env', 'metadata']


class Action(Enum):
    CREATE = 'create'
    UPDATE = 'update'
    INIT = 'init'

    def __str__(self):
        return f'{self.value}'

    def should_patch(self):
        return self is Action.UPDATE


def process_post_result(resp):
    """Prints the result of a token POST"""
    resp_json = resp.json()
    if 'message' in resp_json:
        message = resp_json['message']
        print_info(f'{message}.')
        return

    raise Exception(f'{response_message(resp_json)}')


def post_failed_message(cluster_name, reason):
    """Generates a failed token post message with the given cluster name and reason"""
    return f'Token post {terminal.failed("failed")} on {cluster_name}:\n{terminal.reason(reason)}'


def get_overrides(token_fields_base, token_fields_from_args):
    """Returns true if there are any overrides from the token key args. Handles nested fields args split by a period."""
    overrides = []
    for key_raw, _ in token_fields_from_args.items():
        keys = key_raw.split('.')
        base_ref = token_fields_base
        try:
            for key in keys:
                base_ref = base_ref[key]
            # no KeyError means that the token_fields_base has an existing value corresponding with the arg
            overrides.append(key_raw)
        except KeyError:
            pass
    return overrides


def merge_token_fields_from_args(token_fields_base, token_fields_from_args):
    """Merges token fields from json and token fields from args. Handles nested fields args split by a period."""
    token_fields = {**token_fields_base}
    for key_raw, value in token_fields_from_args.items():
        keys = key_raw.split('.')
        update_in(token_fields, keys, value)
    return token_fields


def create_or_update(cluster, token_name, token_fields, admin_mode, action, fields_from_args_only, output):
    """Creates (or updates) the given token on the given cluster"""
    cluster_name = cluster['name']
    cluster_url = cluster['url']

    existing_token_data, existing_token_etag = get_token(cluster, token_name)
    try:
        print_info(f'Attempting to {action} token {("in ADMIN mode " if admin_mode else "")}'
                   f'{("with dry-run enabled " if output else "")}on {terminal.bold(cluster_name)}...')
        params = {'token': token_name}
        if admin_mode:
            params['update-mode'] = 'admin'
        json_body = existing_token_data if existing_token_data and action.should_patch() else {}
        if fields_from_args_only and action.should_patch():
            json_body = deep_merge(json_body, token_fields)
        else:
            json_body.update(token_fields)
        if output is None:
            headers = {'If-Match': existing_token_etag or ''}
            resp = http_util.post(cluster, 'token', json_body, params=params, headers=headers)
            process_post_result(resp)
        elif output == '-':
            print_info('Token configuration (as json) is:')
            display_data({'json': True}, json_body)
        else:
            output_format = 'json' if output.endswith('.json') else 'yaml'
            output_options = {output_format: True}
            data_format = determine_format(output_options)
            print_info(f'Writing token configuration (as {output_format}) to {output}')
            with open(output, 'w') as out_file:
                data_format.dump(json_body, out_file)
        return 0
    except requests.exceptions.ReadTimeout as rt:
        logging.exception(rt)
        print_info(terminal.failed(
            f'Encountered read timeout with {cluster_name} ({cluster_url}). Your post may have completed.'))
        return 1
    except IOError as ioe:
        logging.exception(ioe)
        reason = f'Cannot connect to {cluster_name} ({cluster_url})'
        message = post_failed_message(cluster_name, reason)
        print_info(f'{message}\n')


def pop_context_override_args(args):
    context_overrides = {}
    for k in list(args.keys()):
        if k.startswith('context.'):
            v = args.pop(k)
            context_overrides[k] = v
    if context_overrides:
        return merge_token_fields_from_args({}, context_overrides)['context']
    else:
        return None


def create_or_update_token(clusters, args, _, enforce_cluster, action):
    """Creates (or updates) a Waiter token"""
    guard_no_cluster(clusters)
    logging.debug('args: %s' % args)
    token_name_from_args = args.pop('token', None)
    json_file = args.pop('json', None)
    yaml_file = args.pop('yaml', None)
    input_file = args.pop('input', None)
    admin_mode = args.pop('admin', None)
    output = args.pop('output', None)
    allow_override = args.pop('override', False)
    context_file = args.pop('context', None)
    context_overrides = pop_context_override_args(args)

    if input_file or json_file or yaml_file:
        token_fields_from_json = load_data({'context_file': context_file,
                                            'context_overrides': context_overrides,
                                            'data': input_file,
                                            'json': json_file,
                                            'yaml': yaml_file})
        fields_from_args_only = False
    else:
        if context_file:
            raise Exception('The --context file can only be used when a data file is specified via '
                            '--input, --json, or --yaml.')
        if context_overrides:
            raise Exception('The --context.xyz overrides can only be used when a data file is specified via '
                            '--input, --json, or --yaml.')
        token_fields_from_json = {}
        fields_from_args_only = True

    token_fields_from_args = args
    overrides = get_overrides(token_fields_from_json, token_fields_from_args)
    if overrides:
        if not allow_override:
            raise Exception(f'You cannot specify the same parameter in both an input file '
                            f'and token field flags at the same time ({", ".join(overrides)}) '
                            f'without specifying the --override flag.')
        else:
            logging.debug(f'Following parameters have specified values in both file and flags: {overrides}')
    token_fields = merge_token_fields_from_args(token_fields_from_json, token_fields_from_args)
    token_name_from_json = token_fields.pop('token', None)
    if token_name_from_args and token_name_from_json:
        if not allow_override:
            raise Exception('You cannot specify the token name both as an argument and in the input file '
                            'without specifying the --override flag.')
        else:
            logging.debug(f'Will use token ({token_name_from_args}) from args and '
                          f'skip token specified in file({token_name_from_json})')

    token_name = token_name_from_args or token_name_from_json
    if not token_name:
        raise Exception('You must specify the token name either as an argument or in an input file via '
                        '--json or --yaml.')

    if len(clusters) > 1:
        default_for_create = [c for c in clusters if c.get('default-for-create', False)]
        num_default_create_clusters = len(default_for_create)
        if num_default_create_clusters == 0:
            raise Exception('You must either specify a cluster via --cluster or set "default-for-create" to true for '
                            'one of your configured clusters.')
        elif num_default_create_clusters > 1:
            raise Exception('You have "default-for-create" set to true for more than one cluster.')
        else:
            query_result = query_token(clusters, token_name)
            if query_result['count'] > 0:
                cluster = get_target_cluster_from_token(clusters, token_name, enforce_cluster)
                logging.debug(f'token already exists in: {cluster}')
            else:
                cluster = default_for_create[0]
    else:
        cluster = clusters[0]

    return create_or_update(cluster, token_name, token_fields, admin_mode, action, fields_from_args_only, output)


def add_arguments(parser):
    """Adds arguments to the given parser"""
    add_token_flags(parser)
    parser.add_argument('token', nargs='?')
    if is_admin_enabled():
        parser.add_argument('--admin', '-a', help='run command in admin mode', action='store_true')
    format_group = parser.add_mutually_exclusive_group()
    format_group.add_argument('--json', help='provide the data in a JSON file', dest='json')
    format_group.add_argument('--yaml', help='provide the data in a YAML file', dest='yaml')
    format_group.add_argument('--input', help='provide the data in a JSON/YAML file', dest='input')
    parser.add_argument('--output', help='outputs the computed token configuration in a JSON/YAML file (or to stdout using -)'
                                         'without performing any token edit operations')
    parser.add_argument('--context', dest='context',
                        help='can be used only when a data file has been provided via --input, --json, or --yaml; '
                             'this JSON/YAML file provides the context variables used '
                             'to render the data file as a template')
    add_override_flags(parser)


def add_token_flags(parser):
    """Adds the "core" token-field flags to the given parser"""
    parser.add_argument('--name', '-n', help='name of service')
    parser.add_argument('--owner', '-o', help='owner of service')
    parser.add_argument('--version', '-v', help='version of service')
    parser.add_argument('--cmd', '-C', help='command to start service')
    parser.add_argument('--cmd-type', '-t', help='command type of service (e.g. "shell")', dest='cmd-type')
    parser.add_argument('--cpus', '-c', help='cpus to reserve for service', type=float)
    parser.add_argument('--mem', '-m', help='memory (in MiB) to reserve for service', type=int)
    parser.add_argument('--ports', help='number of ports to reserve for service', type=int)


def add_override_flags(parser):
    """Adds the arguments override file values flags to the given parser"""
    override_group = parser.add_mutually_exclusive_group(required=False)
    override_group.add_argument('--override', action='store_true', dest='override',
                                help='Allow overriding values in input file with values from CLI arguments. '
                                     'Overriding values is disallowed by default. '
                                     'Adding the --no-override flag explicitly disallows overriding values.')
    override_group.add_argument('--no-override', action='store_false', dest='override', help=argparse.SUPPRESS)


def register_argument_parser(add_parser, action):
    """Calls add_parser for the given sub-command (create or update) and returns the parser"""
    sub_command = str(action)
    return add_parser(sub_command,
                      help=f'{sub_command} token',
                      description=f'{sub_command.capitalize()} a Waiter token. '
                      'In addition to the optional arguments '
                      'explicitly listed below, '
                      'you can optionally provide any Waiter '
                      'token parameter as a flag. For example, '
                      'to specify 10 seconds for the '
                      'grace-period-secs parameter, '
                      'you can pass --grace-period-secs 10. '
                      'You can also provide nested fields separated by a period. For example, '
                      'to specify an environment variable FOO as \"bar\", you can pass --env.FOO \"bar\".')


def possible_int(arg):
    """Attempts to parse arg as an int, returning the string otherwise"""
    try:
        return int(arg)
    except ValueError:
        logging.info(f'failed to parse {arg} as an int, treating it as a string')
        return arg


def possible_float(arg):
    """Attempts to parse arg as a float, returning the string otherwise"""
    try:
        return float(arg)
    except ValueError:
        logging.info(f'failed to parse {arg} as a float, treating it as a string')
        return arg


def add_implicit_arguments(unknown_args, parser):
    """
    Given the list of "unknown" args, dynamically adds proper arguments to
    the subparser, allowing us to support any token parameter as a flag
    """
    num_unknown_args = len(unknown_args)
    for i in range(num_unknown_args):
        arg = unknown_args[i]
        if arg.startswith(("-", "--")):
            arg_dest = arg.lstrip('-')
            if any(arg_dest.startswith(prefix) for prefix in STRING_PARAM_PREFIXES):
                arg_type = None
            elif any(arg.endswith(suffix) for suffix in INT_PARAM_SUFFIXES):
                arg_type = possible_int
            elif any(arg.endswith(suffix) for suffix in FLOAT_PARAM_SUFFIXES):
                arg_type = possible_float
            elif (i + 1) < num_unknown_args and unknown_args[i + 1].lower() in BOOL_STRINGS:
                arg_type = str2bool
            else:
                arg_type = None
            parser.add_argument(arg, dest=arg_dest, type=arg_type)



from functools import partial

import requests

from waiter import terminal, http_util
from waiter.action import ping_token_on_cluster, process_kill_request
from waiter.querying import get_target_cluster_from_token, get_token
from waiter.token_post import post_failed_message, process_post_result
from waiter.util import check_positive, guard_no_cluster, logging, print_info


def _is_token_in_maintenance_mode(token_data):
    return 'maintenance' in token_data


def _get_existing_token_data(clusters, token_name, enforce_cluster):
    guard_no_cluster(clusters)
    cluster = get_target_cluster_from_token(clusters, token_name, enforce_cluster)
    existing_token_data, existing_token_etag = get_token(cluster, token_name)
    return cluster, existing_token_data, existing_token_etag


def _update_token(cluster, token_name, existing_token_etag, body):
    cluster_name = cluster['name']
    cluster_url = cluster['url']
    headers = {'If-Match': existing_token_etag}
    params = {'token': token_name}
    try:
        resp = http_util.post(cluster, 'token', body, params=params, headers=headers)
        process_post_result(resp)
        token_etag = resp.headers.get('ETag', None)
        return 0, token_etag
    except requests.exceptions.ReadTimeout as rt:
        logging.exception(rt)
        print_info(terminal.failed(
            f'Encountered read timeout with {cluster_name} ({cluster_url}). The operation may have completed.'))
        return 1, None
    except IOError as ioe:
        logging.exception(ioe)
        reason = f'Cannot connect to {cluster_name} ({cluster_url})'
        message = post_failed_message(cluster_name, reason)
        print_info(f'{message}\n')
        return 1, None


def check_maintenance(clusters, args, _, enforce_cluster):
    """Checks if a token is in maintenance mode and displays the result. Returns 0 if the token is in maintenance mode
    and returns 1 if the token is NOT in maintenance mode."""
    token_name = args['token']
    _, existing_token_data, existing_token_etag = _get_existing_token_data(clusters, token_name, enforce_cluster)
    maintenance_mode_active = _is_token_in_maintenance_mode(existing_token_data)
    print_info(f'{token_name} is {"" if maintenance_mode_active else "not "}in maintenance mode')
    return 0 if maintenance_mode_active else 1


def start_maintenance(clusters, args, _, enforce_cluster):
    """Sets the token in maintenance mode by updating the token user metadata fields"""
    token_name = args['token']
    timeout_secs = args['timeout']
    kill_services_option = args.pop('kill_services', 'force_kill')
    cluster, existing_token_data, existing_token_etag = _get_existing_token_data(clusters, token_name, enforce_cluster)
    json_body = existing_token_data
    update_doc = {"maintenance": {"message": args['message']}}
    json_body.update(update_doc)
    return_code, token_etag = _update_token(cluster, token_name, existing_token_etag, json_body)
    if return_code == 0:
        print_info(f'Maintenance mode activated for {terminal.bold(token_name)} on cluster '
                   f'{terminal.bold(cluster["name"])}.')
        kill_services = False
        force_flag = False
        if kill_services_option == 'force_kill':
            kill_services = True
            force_flag = True
        elif kill_services_option == 'ask_kill':
            kill_services = True
            force_flag = False
        logging.debug(f"maintenance start will {'' if kill_services else 'not '} "
                      f"{'force ' if force_flag else ''} kill services")
        if kill_services:
            if token_etag:
                success = process_kill_request(clusters, token_name, False, force_flag, timeout_secs, True)
                return 0 if success else 1
            else:
                logging.debug(f'Not killing services for token {token_name} in {cluster} as token ETag is missing.')
                return 1
        else:
            logging.debug(f'Skipped killing services for token {token_name} in {cluster}.')
            return 0
    else:
        return return_code


def stop_maintenance(clusters, args, _, enforce_cluster):
    """Stops maintenance mode for a token by deleting the 'maintenance' user metadata field in the token data"""
    token_name = args['token']
    ping_token = args.pop('ping_token', True)
    timeout = args.pop('timeout', 300)
    wait_for_ping = args.pop('wait', True)
    check_maintenance_mode = args.pop('check', False)
    cluster, existing_token_data, existing_token_etag = _get_existing_token_data(clusters, token_name, enforce_cluster)
    maintenance_mode_active = _is_token_in_maintenance_mode(existing_token_data)
    logging.debug(f'Token {token_name} in {cluster} with etag {existing_token_etag} '
                  f'is {"" if maintenance_mode_active else "not"} in maintenance mode')
    if not maintenance_mode_active and check_maintenance_mode:
        raise Exception("Token is not in maintenance mode")
    if maintenance_mode_active:
        json_body = existing_token_data
        json_body.pop("maintenance")
        return_code, token_etag = _update_token(cluster, token_name, existing_token_etag, json_body)
    else:
        print(f'Token {terminal.bold(token_name)} does not have maintenance mode activated')
        return_code, token_etag = 0, existing_token_etag
    if return_code == 0:
        if ping_token:
            if token_etag:
                success = ping_token_on_cluster(cluster, token_name, timeout, wait_for_ping, token_etag)
                return 0 if success else 1
            else:
                logging.debug(f'Not pinging token {token_name} in {cluster} as token ETag is missing.')
                return 1
        else:
            logging.debug(f'Skipped pinging token {token_name} in {cluster}.')
            return 0
    else:
        return return_code


def maintenance(parser, clusters, args, _, enforce_cluster):
    """Calls the sub action for maintenance command. If no sub action is provided then displays the help message."""
    logging.debug('args: %s' % args)
    sub_func = args.get('sub_func', None)
    if sub_func is None:
        parser.print_help()
        return 0
    else:
        return sub_func(clusters, args, _, enforce_cluster)


def register_check(add_parser):
    """Registers the maintenance check parser"""
    parser = add_parser('check',
                        help='checks if a token is in maintenance mode. Exits with code 0 if the token is in '
                             'maintenance mode and 1 if the token is not in maintenance mode')
    parser.add_argument('token')
    parser.set_defaults(sub_func=check_maintenance)


def register_stop(command_name, add_parser):
    """Registers the maintenance stop parser"""
    parser = add_parser(command_name,
                        help='stop maintenance mode for a token. '
                             'Requests to the token will be handled normally. '
                             'By default, also ping the token to ensure a running service.')
    ping_group = parser.add_mutually_exclusive_group(required=False)
    ping_group.add_argument('--no-ping', action='store_false', dest='ping_token',
                            help='skips pinging the token; pinging the token is enabled by default')
    ping_group.add_argument('--ping', action='store_true', dest='ping_token',
                            help='pings the token after stopping maintenance.')
    parser.add_argument('--timeout', '-t', default=300, help='read timeout (in seconds) for ping request',
                        type=check_positive)
    parser.add_argument('--no-wait', '-n', help='do not wait for ping request to return',
                        dest='wait', action='store_false')
    check_group = parser.add_mutually_exclusive_group(required=False)
    check_group.add_argument('--check', action='store_true', dest='check',
                             help='enforces the check that the token is currently in maintenance mode')
    check_group.add_argument('--no-check', action='store_false', dest='check',
                             help='skips checking maintenance mode; skipping the check is enabled by default')
    parser.add_argument('token')
    parser.set_defaults(sub_func=stop_maintenance)
    return stop_maintenance


def register_start(command_name, add_parser):
    """Registers the maintenance start parser"""
    parser = add_parser(command_name,
                        help='start maintenance mode for a token. '
                             'All requests to this token will begin to receive a 503 response. '
                             "By default, also kill the token's currently running services.")
    kill_group = parser.add_mutually_exclusive_group(required=False)
    kill_group.add_argument('--ask-kill', action='store_const', const='ask_kill', dest='kill_services',
                            help="Ask before killing if there are multiple running services accessed by the token.")
    kill_group.add_argument('--force-kill', action='store_const', const='force_kill', dest='kill_services',
                            help="Force kill the token's currently running services. "
                                 "Killing the token's services is enabled by default.")
    kill_group.add_argument('--no-kill', action='store_const', const='no_kill', dest='kill_services',
                            help="Skip killing the token's currently running services.")
    parser.add_argument('--timeout', '-t', default=10, help='timeout (in seconds) for service kill requests.',
                        type=check_positive)
    parser.add_argument('token')
    parser.add_argument('message',
                        help='Your message will be provided in a 503 response for requests to the token. '
                             'The message cannot be more than 512 characters.')
    parser.set_defaults(sub_func=start_maintenance)
    return start_maintenance


def register(add_parser):
    """Adds this sub-command's parser and returns the action function"""
    parser = add_parser('maintenance',
                        help='manage maintenance mode for a token',
                        description='Manage maintenance mode for a Waiter token.')
    subparsers = parser.add_subparsers()
    register_check(subparsers.add_parser)
    register_start('start', subparsers.add_parser)
    register_stop('stop', subparsers.add_parser)
    return partial(maintenance, parser)



import json
import logging
import os
import sys

from waiter.util import deep_merge, load_json_file

# Base locations to check for configuration files, relative to the executable.
# Always tries to load these in.
BASE_CONFIG_PATHS = ['.waiter.json',
                     '../.waiter.json',
                     '../config/.waiter.json']

# Additional locations to check for configuration files if one isn't given on the command line
ADDITIONAL_CONFIG_PATHS = ['.waiter.json',
                           os.path.expanduser('~/.waiter.json')]

DEFAULT_CONFIG = {'http': {'retries': 2,
                           'connect-timeout': 3.05,
                           'read-timeout': 20},
                  'metrics': {'disabled': True,
                              'max-retries': 2,
                              'timeout': 0.15}}


def __load_first_json_file(paths):
    """Returns the contents of the first parseable JSON file in a list of paths."""
    if paths is None:
        paths = []
    contents = ((os.path.abspath(p), load_json_file(os.path.abspath(p))) for p in paths if p)
    return next(((p, c) for p, c in contents if c), (None, None))


def __load_base_config():
    """Loads the base configuration map."""
    base_dir = os.path.dirname(os.path.abspath(sys.argv[0]))
    paths = [os.path.join(base_dir, path) for path in BASE_CONFIG_PATHS]
    _, config = __load_first_json_file(paths)
    return config


def __load_local_config(config_path):
    """
    Loads the configuration map, using the provided config_path if not None,
    otherwise, searching the additional config paths for a valid JSON config file
    """
    if config_path:
        if os.path.isfile(config_path):
            with open(config_path) as json_file:
                config = json.load(json_file)
        else:
            raise Exception(f'The configuration path specified ({config_path}) is not valid.')
    else:
        config_path, config = __load_first_json_file(ADDITIONAL_CONFIG_PATHS)

    return config_path, config


def __prepend_additional_clusters(config):
    """Prepend user's local additional-clusters to the base config clusters."""
    if 'additional-clusters' in config:
        clusters = config.get('clusters', [])
        additional_clusters = config.get('additional-clusters', [])
        config['clusters'] = additional_clusters + clusters
        del config['additional-clusters']


def load_config_with_defaults(config_path=None):
    """Loads the configuration map to use, merging in the defaults"""
    base_config = __load_base_config()
    base_config = base_config or {}
    base_config = deep_merge(DEFAULT_CONFIG, base_config)
    config_path, config = __load_local_config(config_path)
    config = config or {}
    config = deep_merge(base_config, config)
    __prepend_additional_clusters(config)
    logging.debug(f'using configuration: {config}')
    return config



import json
import logging
import os
import string
import sys

import yaml


class DataFormat:
    def __str__(self):
        """Returns the name of the format in upper case."""
        return self.name().upper()

    def name(self):
        """Returns the name of the format."""
        raise NotImplementedError('Method has not been implemented!')

    def parse(self, data):
        """Returns the parsed data."""
        raise NotImplementedError('Method has not been implemented!')

    def dump(self, out_data, out_file=None):
        """
        If the file is provided, writes the string representation of the data to the file.
        Else it returns the string representation of the data in the format.
        """
        raise NotImplementedError('Method has not been implemented!')


class JsonDataFormat(DataFormat):
    def name(self):
        return 'json'

    def parse(self, data):
        try:
            logging.debug(f'parsing input data as json')
            content = json.loads(data)
            return content
        except Exception:
            raise ValueError('Malformed JSON in input.')

    def dump(self, out_data, out_file=None):
        if out_file:
            json.dump(out_data, out_file, indent=2, sort_keys=True)
        else:
            return json.dumps(out_data, indent=2, sort_keys=True)


class YamlDataFormat(DataFormat):
    def name(self):
        return 'yaml'

    def parse(self, data):
        try:
            logging.debug(f'parsing input data as yaml')
            content = yaml.safe_load(data)
            return content
        except Exception:
            raise ValueError('Malformed YAML in input.')

    def dump(self, out_data, out_file=None):
        if out_file:
            yaml.safe_dump(out_data, out_file)
        else:
            return yaml.safe_dump(out_data)


JSON = JsonDataFormat()
YAML = YamlDataFormat()


class AnySupportedFormat(DataFormat):
    def name(self):
        return 'data'

    def parse(self, data):
        content = None
        for input_format in [JSON, YAML]:
            try:
                logging.debug(f'attempting to parse input data as {input_format}')
                content = input_format.parse(data)
                logging.debug(f'successfully parsed input data as {input_format}')
                continue
            except Exception:
                logging.debug(f'error parsing input data as {input_format}')
        if content is None:
            raise ValueError('Malformed data in input.')
        return content

    def dump(self, out_data, out_file=None):
        raise NotImplementedError('Data format needs to be specified explicitly!')


ANY_FORMAT = AnySupportedFormat()


def validate_options(options):
    """Validates whether unique file format is specified."""
    as_json = options.get(JSON.name())
    as_yaml = options.get(YAML.name())
    if as_json and as_yaml:
        raise Exception(f'JSON and YAML mode cannot be used simultaneously!')


def determine_format(options):
    """
    Determines whether the configured format is YAML or JSON.
    JSON is the default format if YAML is not explicitly enabled.
    """
    validate_options(options)
    as_yaml = options.get(YAML.name())
    data_format = YAML if as_yaml else JSON
    return data_format


def read_from_standard_input(input_format):
    """Prompts for and then reads token JSON from stdin"""
    logging.debug(f'expecting {input_format} input from stdin')
    print(f'Enter the raw {input_format} (press Ctrl+D on a blank line to submit)', file=sys.stderr)
    data = sys.stdin.read()
    return data


def load_file(path):
    """Loads the file"""
    content = None

    if os.path.isfile(path):
        with open(path) as data_file:
            try:
                logging.debug(f'attempting to load content from {path}')
                content = data_file.read()
            except Exception:
                logging.exception(f'encountered exception when loading content from {path}')
    else:
        logging.info(f'{path} is not a file')

    return content


def load_data(options):
    """
    Decode a JSON/YAML formatted file.
    Data must be a dict of attributes.
    Throws a ValueError if there is a problem parsing the data.
    """
    input_format = ANY_FORMAT if options.get('data') else determine_format(options)
    input_file = options.get(input_format.name())

    if input_file == '-':
        logging.debug(f'reading {input_format} input from stdin')
        content = read_from_standard_input(input_format)
        if not content:
            raise Exception(f'Unable to load {input_format} from stdin.')
    else:
        logging.debug(f'reading {input_format} input from {input_file}')
        content = load_file(input_file)
        if not content:
            raise Exception(f'Unable to load {input_format} from {input_file}.')

        context_dict = {}
        context_provided = False
        context_file = options.get('context_file')
        if context_file:
            context_provided = True
            logging.debug(f'reading context from {context_file}')
            context_content = load_file(context_file)
            if not context_content:
                raise Exception(f'Unable to load context from {context_file}.')

            context_file_obj = YAML.parse(context_content)
            if not isinstance(context_file_obj, dict):
                raise Exception(f'Provided context file must evaluate to a dictionary, instead it is {context_file_obj}')
            context_dict.update(context_file_obj)

        context_overrides = options.get('context_overrides')
        if context_overrides:
            context_provided = True
            logging.debug(f'merging additional context {context_overrides}')
            context_dict.update(context_overrides)

        if context_provided:
            try:
                logging.debug(f'applying string templating to input using context {context_dict}')
                string_template = string.Template(content)
                content = string_template.substitute(context_dict)
            except Exception as ex:
                message = f'missing variable {ex}' if isinstance(ex, KeyError) else str(ex)
                raise Exception(f'Error when processing template: {message}')

    content = input_format.parse(content)
    if type(content) is dict:
        return content
    else:
        raise ValueError(f'Input {input_format} must be a dictionary of attributes.')


def display_data(options, data):
    """Display data as JSON/YAML format to standard output."""
    input_format = determine_format(options)
    result = input_format.dump(data)
    if result:
        print(result)



from waiter.action import process_kill_request, process_ping_request
from waiter.util import guard_no_cluster, check_positive


def kill(clusters, args, _, __):
    """Kills the service(s) using the given token name."""
    guard_no_cluster(clusters)
    token_name_or_service_id = args.get('token-or-service-id')
    is_service_id = args.get('is-service-id', False)
    force_flag = args.get('force', False)
    timeout_secs = args['timeout']
    ping_token_name_or_service_id = args.get('ping_token', False)

    success = process_kill_request(clusters, token_name_or_service_id, is_service_id, force_flag, timeout_secs)
    if success and ping_token_name_or_service_id:
        ping_timeout_secs = args.get('ping_timeout')
        ping_wait = args.get('ping_wait', True)
        ping_success = process_ping_request(clusters, token_name_or_service_id, is_service_id,
                                            ping_timeout_secs, ping_wait)
        return 0 if ping_success else 1
    else:
        return 0 if success else 1


def register(add_parser):
    """Adds this sub-command's parser and returns the action function"""
    parser = add_parser('kill', help='kill services')
    parser.add_argument('token-or-service-id')
    parser.add_argument('--force', '-f', help='kill all services, never prompt', dest='force', action='store_true')
    parser.add_argument('--service-id', '-s', help='kill by service id instead of token',
                        dest='is-service-id', action='store_true')
    parser.add_argument('--timeout', '-t', help='timeout (in seconds) for kill to complete',
                        type=check_positive, default=30)

    ping_group = parser.add_mutually_exclusive_group(required=False)
    ping_group.add_argument('--no-ping', action='store_false', dest='ping_token',
                            help='skips pinging the token/service; pinging the token is disabled by default')
    ping_group.add_argument('--ping', action='store_true', dest='ping_token',
                            help='pings the token/service after killing service(s).')
    parser.set_defaults(ping_token=False)
    parser.add_argument('--ping-timeout', default=300, dest='ping_timeout',
                        help='read timeout (in seconds) for ping request (default=300)', type=check_positive)
    ping_wait_group = parser.add_mutually_exclusive_group(required=False)
    ping_wait_group.add_argument('--no-ping-wait', action='store_false', dest='ping_wait',
                                 help='do not wait for ping request to return')
    ping_wait_group.add_argument('--ping-wait', action='store_true', dest='ping_wait',
                                 help='wait for ping request to return (default)')
    parser.set_defaults(ping_wait=True)

    return kill



import argparse
import json
import logging
import os
import sys
import time
from datetime import datetime, timedelta

from waiter import terminal

TRUE_STRINGS = ('yes', 'true', 'y')
FALSE_STRINGS = ('no', 'false', 'n')


def get_in(obj, keys, default_value=None):
    """Given a list of keys and a value, return the nested value in the object or the default"""
    cur_node = obj
    for key in keys:
        if key in cur_node:
            cur_node = cur_node[key]
        else:
            return default_value
    return cur_node


def update_in(obj, keys, value):
    """Given a list of keys and a value, return a new object with that value set"""
    cur_node = obj
    for key in keys[:-1]:
        cur_node = cur_node.setdefault(key, {})
    cur_node[keys[-1]] = value


def deep_merge(a, b):
    """Merges a and b, letting b win if there is a conflict"""
    merged = a.copy()
    for key in b:
        b_value = b[key]
        merged[key] = b_value
        if key in a:
            a_value = a[key]
            if isinstance(a_value, dict) and isinstance(b_value, dict):
                merged[key] = deep_merge(a_value, b_value)
    return merged


def current_user():
    """Returns the value of the USER environment variable"""
    return os.environ['USER']


def print_error(text):
    """Prints text to stderr, colored as a failure"""
    print(terminal.failed(text), file=sys.stderr)


def print_info(text, end='\n'):
    """Prints text to stdout"""
    print(text, flush=True, end=end)


def guard_no_cluster(clusters):
    """Throws if no clusters have been specified, either via configuration or via the command line"""
    if not clusters:
        raise Exception('You must specify at least one cluster.')


def str2bool(v):
    """Converts the given string to a boolean, or returns None"""
    if v.lower() in TRUE_STRINGS:
        return True
    elif v.lower() in FALSE_STRINGS:
        return False
    else:
        return None


def response_message(resp_json):
    """Pulls the error message out of a Waiter response"""
    if 'waiter-error' in resp_json and 'message' in resp_json['waiter-error']:
        message = resp_json['waiter-error']['message']
        if not message.endswith('.'):
            message = f'{message}.'
    else:
        message = 'Encountered unexpected error.'
    return message


def wait_until(pred, timeout=30, interval=5):
    """
    Wait, retrying a predicate until it is True, or the
    timeout value has been exceeded.
    """
    if timeout:
        finish = datetime.now() + timedelta(seconds=timeout)
    else:
        finish = None

    while True:
        result = pred()

        if result:
            break

        if finish and datetime.now() >= finish:
            break

        time.sleep(interval)

    return result


def check_positive(value):
    """Checks that the given value is a positive integer"""
    try:
        integer = int(value)
    except:
        raise argparse.ArgumentTypeError(f'{value} is not an integer')
    if integer <= 0:
        raise argparse.ArgumentTypeError(f'{value} is not a positive integer')
    return integer


def load_json_file(path):
    """Decode a JSON formatted file."""
    content = None

    if os.path.isfile(path):
        with open(path) as json_file:
            try:
                logging.debug(f'attempting to load json from {path}')
                content = json.load(json_file)
            except Exception:
                logging.exception(f'encountered exception when loading json from {path}')
    else:
        logging.info(f'{path} is not a file')

    return content


def is_service_current(service, current_token_etag, token_name):
    """Returns True if any of the given service's source tokens is the current token"""
    is_current = any(source['version'] == current_token_etag and source['token'] == token_name
                     for sources in service['source-tokens']
                     for source in sources)
    return is_current


def is_admin_enabled():
    """Returns True if current user is an admin"""
    return str2bool(os.getenv('WAITER_ADMIN', default=FALSE_STRINGS[0]))



import concurrent
import logging
import os
from concurrent import futures

from waiter import http_util, terminal


def query_across_clusters(clusters, query_fn):
    """Attempts to query entities from the given clusters."""
    count = 0
    all_entities = {'clusters': {}}
    max_workers = os.cpu_count()
    logging.debug('querying with max workers = %s' % max_workers)
    with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
        future_to_cluster = {query_fn(c, executor): c for c in clusters}
        for future, cluster in future_to_cluster.items():
            entities = future.result()
            cluster_count = entities['count']
            if cluster_count > 0:
                all_entities['clusters'][cluster['name']] = entities
                count += cluster_count
    all_entities['count'] = count
    return all_entities


def get_token(cluster, token_name, include=None):
    """Gets the token with the given name from the given cluster"""
    params = {'token': token_name}
    if include:
        params['include'] = include
    token_data, headers = http_util.make_data_request(cluster, lambda: http_util.get(cluster, 'token', params=params))
    etag = headers.get('ETag', None)
    return token_data, etag


def no_data_message(clusters):
    """Returns a message indicating that no data was found in the given clusters"""
    clusters_text = ' / '.join([c['name'] for c in clusters])
    message = terminal.failed(f'No matching data found in {clusters_text}.')
    message = f'{message}\nDo you need to add another cluster to your configuration?'
    return message


def print_no_data(clusters):
    """Prints a message indicating that no data was found in the given clusters"""
    print(no_data_message(clusters))


def print_no_services(clusters, token):
    """Prints a message that there were no services found for a token"""
    clusters_text = ' / '.join([terminal.bold(c['name']) for c in clusters])
    print(f'There are no services using token {terminal.bold(token)} in {clusters_text}.')


def print_no_instances(service):
    """Prints a message that there are no relevant instances for the service"""
    print(f'There are no relevant instances using service id {terminal.bold(service)}.')
    print(f'Check the --include flags for active, failed, and killed instances.')


def get_token_on_cluster(cluster, token_name, include_services=False, include_deleted=False):
    """Gets the token with the given name on the given cluster"""
    include_flags = ['metadata']
    if include_deleted:
        include_flags.append('deleted')
    token_data, token_etag = get_token(cluster, token_name, include=include_flags)
    if token_data:
        data = {'count': 1, 'token': token_data, 'etag': token_etag}
        if include_services:
            data['services'] = get_services_using_token(cluster, token_name)
        return data
    else:
        logging.info(f'Unable to retrieve token information on {cluster["name"]} ({cluster["url"]}).')
        return {'count': 0}


def query_token(clusters, token, include_services=False, include_deleted=False):
    """
    Uses query_across_clusters to make the token
    requests in parallel across the given clusters
    """

    def submit(cluster, executor):
        return executor.submit(get_token_on_cluster, cluster, token, include_services, include_deleted)

    return query_across_clusters(clusters, submit)


def get_service(cluster, service_id):
    """Retrieves the service with the given service id"""
    params = {'effective-parameters': True}
    endpoint = f'/apps/{service_id}'
    service, _ = http_util.make_data_request(cluster, lambda: http_util.get(cluster, endpoint, params=params))
    return service


def get_service_on_cluster(cluster, service_id):
    """Gets the service with the given service id on the given cluster"""
    service = get_service(cluster, service_id)
    if service:
        return {'count': 1, 'service': service}
    else:
        return {'count': 0}
                     
                     
def get_services_using_token(cluster, token_name):
    """Retrieves all services that are using the token"""
    params = {'effective-parameters': 'true',
              'token': token_name}
    services, _ = http_util.make_data_request(cluster, lambda: http_util.get(cluster, 'apps', params=params))
    return services


def get_services_on_cluster(cluster, token_name):
    """Gets the service(s) using the given token name on the given cluster"""
    services = get_services_using_token(cluster, token_name)
    if services:
        return {'count': len(services), 'services': services}
    else:
        return {'count': 0}


def query_service(clusters, service_id):
    """
    Uses query_across_clusters to make the service
    requests in parallel across the given clusters
    """
    return query_across_clusters(
        clusters,
        lambda cluster, executor: executor.submit(get_service_on_cluster, cluster, service_id))


def query_services(clusters, token_name):
    """
    Uses query_across_clusters to make the service
    requests in parallel across the given clusters
    """
    return query_across_clusters(
        clusters,
        lambda cluster, executor: executor.submit(get_services_on_cluster, cluster, token_name))


def get_tokens(cluster, user):
    """Gets the tokens owned by the given user from the given cluster"""
    params = {'owner': user, 'include': 'metadata'}
    tokens, _ = http_util.make_data_request(cluster, lambda: http_util.get(cluster, 'tokens', params=params))
    return tokens


def get_tokens_on_cluster(cluster, user):
    """Gets the tokens owned by the given user on the given cluster"""
    tokens = get_tokens(cluster, user)
    if tokens:
        data = {'count': len(tokens), 'tokens': tokens}
        return data
    else:
        logging.info(f'Unable to retrieve token information on {cluster["name"]} ({cluster["url"]}).')
        return {'count': 0}


def query_tokens(clusters, user):
    """
    Uses query_across_clusters to make the token
    requests in parallel across the given clusters
    """
    return query_across_clusters(
        clusters,
        lambda cluster, executor: executor.submit(get_tokens_on_cluster, cluster, user))


def _get_latest_cluster(clusters, query_result):
    """
    :param clusters: list of local cluster configs from the configuration file
    :param query_result: value from query_token function
    :return: Finds latest token configuration from the query_result. Gets the cluster that is configured in the
     token description and returns a local cluster who's server side name matches the one specified in the token.
     If the token's cluster does not exist in one of the local cluster configurations then an Exception is raised.
    """
    token_descriptions = list(query_result['clusters'].values())
    token_result = max(token_descriptions, key=lambda token: token['token']['last-update-time'])
    if token_result['token'].get('deleted', False):
        return None
    cluster_name_goal = token_result['token']['cluster'].upper()
    provided_cluster_names = set()
    for c in clusters:
        cluster_settings, _ = http_util.make_data_request(c, lambda: http_util.get(c, '/settings'))
        cluster_config_name = cluster_settings['cluster-config']['name'].upper()
        cluster_local_config_name = c['name'].upper()
        provided_cluster_names.add(cluster_config_name)
        if cluster_name_goal in [cluster_config_name, cluster_local_config_name]:
            return c
    raise Exception(f'The token is configured in cluster {cluster_name_goal}, which is not provided.' +
                    f' The following clusters were provided: {", ".join(provided_cluster_names)}.')


def get_target_cluster_from_token(clusters, token_name, enforce_cluster):
    """
    :param clusters: list of local cluster configs from the configuration file
    :param token_name: string name of token
    :param enforce_cluster: boolean describing if cluster was explicitly specified as an cli argument
    :return: Return the target cluster config for various token operations
    """
    query_result = query_token(clusters, token_name, include_deleted=True)
    if query_result["count"] == 0:
        raise Exception('The token does not exist. You must create it first.')
    elif enforce_cluster:
        logging.debug(f'Forcing cluster {clusters[0]} as the target_cluster')
        return clusters[0]
    else:
        sync_group_count = 0
        sync_groups_set = set()
        cluster_names = set()
        for cluster in list(query_result['clusters'].keys()):
            cluster_config = next(c for c in clusters if c['name'] == cluster)
            sync_group = cluster_config.get('sync-group', False)
            # consider clusters that don't have a configured sync-group as in their own unique group
            if not sync_group:
                sync_group = sync_group_count
                sync_group_count += 1
            sync_groups_set.add(sync_group)
            cluster_names.add(cluster)
        if len(sync_groups_set) > 1:
            raise Exception('Could not infer the target cluster for this operation because there are multiple cluster '
                            f'groups that contain a description for this token: groups-{sync_groups_set} '
                            f'clusters-{cluster_names}.'
                            '\nConsider specifying with the --cluster flag which cluster you are targeting.')
        return _get_latest_cluster(clusters, query_result)


def get_service_id_from_instance_id(instance_id):
    """Extracts the service_id from the instance_id. instance_ids begin with a service_id followed by a period"""
    return instance_id.split('.')[0]
